import os
import unittest

from sqlalchemy.engine import create_engine
from sqlalchemy.orm.session import Session

from geokrety_api_models.base import Base


def setup_module():
    global transaction, connection, engine

    # Connect to the database and create the schema within a transaction
    engine = create_engine(os.environ.get('TEST_DATABASE_URL',
                                          'mysql+mysqldb://root@127.0.0.1/geokrety_unittest?charset=utf8mb4'))
    connection = engine.connect()
    transaction = connection.begin()
    Base.metadata.create_all(connection)
    # If you want to insert fixtures to the DB, do it here


def teardown_module():
    # Roll back the top level transaction and disconnect from the database
    transaction.rollback()
    connection.close()
    engine.dispose()


class DatabaseTest(unittest.TestCase):
    def setUp(self):
        self.__transaction = connection.begin_nested()
        self.session = Session(connection)

    def tearDown(self):
        self.session.close()
        self.__transaction.rollback()



# -*- coding: utf-8 -*-

import datetime
import random

import phpass
from sqlalchemy import (Boolean, Column, DateTime, Integer, String, event,
                        inspect)
from sqlalchemy.dialects.mysql import DOUBLE, INTEGER
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import bleach
import characterentities

from .base import Base


def random_0_23():
    return random.randrange(0, 23)


class User(Base):
    __tablename__ = 'gk-users'

    id = Column(
        'userid',
        Integer,
        primary_key=True,
        key='id',
    )
    _name = Column(
        'user',
        String(80),
        key='name',
        nullable=False,
        unique=True,
    )
    is_admin = Column(
        'is_admin',
        Boolean,
        key='is_admin',
        nullable=True,
        default=False,
    )
    _password = Column(
        'haslo2',
        String(120),
        key='password',
        nullable=False,
    )
    email = Column(
        String(150),
        nullable=False,
        unique=True,
    )
    daily_mails = Column(
        'wysylacmaile',
        Boolean,
        key='daily_news',
        nullable=False,
        default=False,
    )
    ip = Column(
        String(39),
        nullable=True,
        default=None,
    )
    language = Column(
        'lang',
        String(2),
        key='language',
        nullable=False,
        default="",
    )
    latitude = Column(
        'lat',
        DOUBLE(precision=8, scale=5),
        key='latitude',
        nullable=True,
        default=None,
    )
    longitude = Column(
        'lon',
        DOUBLE(precision=8, scale=5),
        key='longitude',
        nullable=True,
        default=None,
    )
    observation_radius = Column(
        'promien',
        INTEGER(unsigned=True),
        key='observation_radius',
        default=0,
    )
    country = Column(
        String(3),
        nullable=True,
        default=None,
    )
    hour = Column(
        'godzina',
        Integer,
        key='hour',
        default=random_0_23,
    )
    statpic_id = Column(
        'statpic',
        Integer,
        key='statpic_id',
        default=1,
    )
    join_datetime = Column(
        'joined',
        DateTime,
        key='join_datetime',
        default=datetime.datetime.utcnow,
    )
    last_mail_datetime = Column(
        'ostatni_mail',
        DateTime,
        nullable=True,
        key='last_mail_datetime',
        default=None,
    )
    last_login_datetime = Column(
        'ostatni_login',
        DateTime,
        nullable=True,
        key='last_login_datetime',
        default=None,
    )
    last_update_datetime = Column(
        'timestamp',
        DateTime,
        key='last_update_datetime',
        default=datetime.datetime.utcnow,
    )
    secid = Column(
        String(128),
    )

    news = relationship(
        'News',
        backref="author",
        cascade="all,delete",
    )
    news_comments = relationship(
        'NewsComment',
        backref="author",
        cascade="all,delete",
    )
    news_subscriptions = relationship(
        'NewsSubscription',
        backref="user",
        cascade="all,delete",
    )

    @hybrid_property
    def password(self):
        """
        Hybrid property for password
        :return:
        """
        return self._password

    @password.setter
    def password(self, password):
        """
        Setter for _password, saves hashed password, salt and reset_password string
        :param password:
        :return:
        """
        t_hasher = phpass.PasswordHash(11, False)
        self._password = t_hasher.hash_password(
            password.encode('utf-8') + "DDD"  # app.config['PASSWORD_HASH_SALT']
        )

    @password.expression
    def password(cls):
        return cls._password

    @hybrid_property
    def name(self):
        return characterentities.decode(self._name)

    @name.setter
    def name(self, name):
        name_clean = bleach.clean(name, tags=[], strip=True)
        self._name = characterentities.decode(name_clean).strip()

    @name.expression
    def name(cls):
        return cls._name


@event.listens_for(User, 'init')
def receive_init(target, args, kwargs):
    target.secid = ''.join(random.choice('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ') for i in range(84))  # TODO
    # target.ip = request.remote_addr


MONITORED_ATTRIBUTES = [
    '_name',
    'is_admin',
    '_password',
    'email',
    'daily_mails',
    'ip',
    'language',
    'latitude',
    'longitude',
    'observation_radius',
    'country',
    'hour',
    'statpic_id',
    'last_mail_datetime',
    'last_login_datetime',
    'last_update_datetime',
    'secid',
]


def _has_changes_that_need_event(instance):
    instance_attrs = inspect(instance).attrs
    for attribute in MONITORED_ATTRIBUTES:
        if hasattr(instance_attrs, attribute) and \
                getattr(instance_attrs, attribute).history.has_changes():
            return True


if __name__ == '__main__':
    # Check
    user = User()



# -*- coding: utf-8 -*-

import datetime

from sqlalchemy import (Column, DateTime, ForeignKey, Integer, String, event,
                        inspect)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import bleach
import characterentities
from geokrety_api_exceptions.json_api import GKUnprocessableEntity

from .base import Base
from .utilities.const import (MOVE_COMMENT_TYPE_MISSING, MOVE_TYPE_ARCHIVED,
                              MOVE_TYPE_DROPPED, MOVE_TYPE_SEEN,
                              MOVE_TYPES_TEXT)


class MoveComment(Base):
    __tablename__ = 'gk-ruchy-comments'

    id = Column(
        'comment_id',
        Integer,
        primary_key=True,
        key='id',
    )

    move_id = Column(
        'ruch_id',
        Integer,
        ForeignKey('gk-ruchy.id'),
        key='move_id',
        nullable=False,
    )

    move = relationship(
        "Move",
        foreign_keys=[move_id],
        backref="comments",
    )

    author_id = Column(
        'user_id',
        Integer,
        ForeignKey('gk-users.id', name='fk_move_comment_author'),
        key='author_id',
        nullable=False,
    )

    author = relationship(
        "User",
        foreign_keys=[author_id],
        backref="moves_comments",
    )

    geokret_id = Column(
        'kret_id',
        Integer,
        ForeignKey('gk-geokrety.id', name='fk_move_comment_geokret'),
        key='geokret_id',
        nullable=False,
    )

    created_on_datetime = Column(
        'data_dodania',
        DateTime,
        key='created_on_datetime',
        nullable=False,
        default=datetime.datetime.utcnow,
    )

    updated_on_datetime = Column(
        'timestamp',
        DateTime,
        key='updated_on_datetime',
        nullable=False,
        default=datetime.datetime.utcnow,
    )

    _comment = Column(
        'comment',
        String(1000),
        nullable=False,
    )

    type = Column(
        Integer,
        nullable=False,
        default=0,
    )

    @hybrid_property
    def comment(self):
        return characterentities.decode(self._comment)

    @comment.setter
    def comment(self, comment):
        # Drop all html tags
        comment_clean = bleach.clean(comment, strip=True)
        # Strip spaces
        self._comment = characterentities.decode(comment_clean).strip()

    @comment.expression
    def comment(cls):
        return cls._comment

    def is_missing_authorized(self):
        if self.type != MOVE_COMMENT_TYPE_MISSING:
            return True
        if self.move.geokret.last_position is not None and \
                self.move.geokret.last_position.id != self.move.id:
            raise GKUnprocessableEntity("Cannot declare 'missing' on old moves",
                                        {'pointer': '/data/relationships/move/data'})
        if self.move.type in [MOVE_TYPE_DROPPED,
                              MOVE_TYPE_SEEN,
                              MOVE_TYPE_ARCHIVED]:
            return True
        raise GKUnprocessableEntity("Cannot declare 'missing' on move_type %s" %
                                    MOVE_TYPES_TEXT[self.move.type],
                                    {'pointer': '/data/relationships/move/data'})


@event.listens_for(MoveComment, 'before_insert')
def before_insert_listener(mapper, connection, target):
    target.geokret_id = target.move.geokret.id
    target.is_missing_authorized()


@event.listens_for(MoveComment, "before_update")
def before_update(mapper, connection, target):
    target.is_missing_authorized()


def _has_changes_that_need_recompute(instance):
    if inspect(instance).attrs.type.history.has_changes() or \
            inspect(instance).attrs.move.history.has_changes():
        return True



# -*- coding: utf-8 -*-

import datetime

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import bleach
import characterentities

from .base import Base


class News(Base):
    __tablename__ = 'gk-news'

    id = Column(
        'news_id',
        Integer,
        primary_key=True,
        key='id'
    )
    _title = Column(
        'tytul',
        String(50),
        key='title',
        nullable=False
    )
    _content = Column(
        'tresc',
        Text,
        key='content',
        nullable=False
    )
    _username = Column(
        'who',
        String(80),
        key='username',
        nullable=True
    )
    author_id = Column(
        'userid',
        Integer,
        ForeignKey('gk-users.id'),
        key='author_id',
        nullable=False,
    )
    comments_count = Column(
        'komentarze',
        Integer,
        key='comments_count',
        default=0
    )
    last_comment_datetime = Column(
        'ostatni_komentarz',
        DateTime,
        key='last_comment_datetime',
        nullable=False,
        default="0000-00-00T00:00:00"
    )
    created_on_datetime = Column(
        'date',
        DateTime,
        key='created_on_datetime',
        default=datetime.datetime.utcnow
    )
    czas_postu = Column(
        DateTime,
        default="0000-00-00T00:00:00"
    )

    # author = relationship('User', backref=backref('news'))
    news_comments = relationship(
        'NewsComment',
        backref="news",
        cascade="all,delete"
    )
    subscriptions = relationship(
        'NewsSubscription',
        backref="news",
        cascade="all,delete"
    )

    @hybrid_property
    def title(self):
        return characterentities.decode(self._title)

    @title.setter
    def title(self, title):
        title_clean = bleach.clean(title, tags=[], strip=True)
        self._title = characterentities.decode(title_clean).strip()

    @title.expression
    def title(cls):
        return cls._title

    @hybrid_property
    def content(self):
        return characterentities.decode(self._content)

    @content.setter
    def content(self, content):
        # Drop all unallowed html tags
        content_clean = bleach.clean(content, strip=True)
        # Strip spaces
        self._content = characterentities.decode(content_clean).strip()

    @content.expression
    def content(cls):
        return cls._content

    @hybrid_property
    def username(self):
        if self._username:
            return characterentities.decode(self._username)

    @username.setter
    def username(self, username):
        username_clean = bleach.clean(username, tags=[], strip=True)
        self._username = characterentities.decode(username_clean).strip()

    @username.expression
    def username(cls):
        return cls._username



import geopy.distance
import requests

from ..geokret import Geokret
from ..move import Move
from .const import (MOVE_COMMENT_TYPE_COMMENT, MOVE_COMMENT_TYPE_MISSING,
                    MOVE_TYPE_ARCHIVED, MOVE_TYPE_COMMENT, MOVE_TYPE_DIPPED,
                    MOVE_TYPE_DROPPED, MOVE_TYPE_GRABBED, MOVE_TYPE_SEEN)


def update_geokret_and_moves(session, geokrety, moves=None):
    """ Recompute all values
    """
    if moves is None:
        moves = []

    if not isinstance(geokrety, list):
        geokrety = [geokrety]
    if not isinstance(moves, list):
        moves = [moves]

    for move_id in moves:
        update_move_comments_type(move_id)
        update_move_country_and_altitude(move_id)

    for geokret_id in geokrety:
        # Enhance Move content
        update_move_distances(geokret_id)
        # Enhance GeoKret content
        update_geokret_total_moves_count(geokret_id)
        update_geokret_holder(geokret_id)
        update_geokret_missing(geokret_id)

    try:
        session.commit()
    except Exception as e:  # pragma: no cover
        session.rollback()
        raise AssertionError("update_geokret_and_moves error: " + str(e))

    # TODO Generate static files
    # * gpx
    # * csv
    # * geojson
    # * statpic owner
    # * statpic mover
    # *


def update_move_distances(geokret_id):
    """ Recompute and update all moves distances for a GeoKret
    """
    moves = Move.query.filter(Move.geokret_id == geokret_id).order_by(Move.moved_on_datetime.asc())
    geokret = Geokret.query.get(geokret_id)

    last = None
    total_distance = 0
    for move in moves:
        if move.type in (MOVE_TYPE_DROPPED, MOVE_TYPE_SEEN, MOVE_TYPE_ARCHIVED):
            geokret.last_position = move
        elif move.type in (MOVE_TYPE_GRABBED, MOVE_TYPE_DIPPED):
            geokret.last_position = None

        geokret.last_move = move
        if move.latitude is None:
            continue
        if last is None:
            last = move
            continue

        distance = geopy.distance.distance((last.latitude, last.longitude), (move.latitude, move.longitude)).km
        move.distance = int(round(distance))
        total_distance += move.distance
        last = move
    geokret.distance = total_distance


def update_move_country_and_altitude(move_id):
    """ Obtain and update country and altitude of a move
    """
    move = Move.query.get(move_id)

    if move.latitude is not None and move.longitude is not None:
        response = requests.get(
            'https://geo.kumy.org/api/getCountry?lat={}&lon={}'.format(move.latitude, move.longitude))
        if response.ok:
            move.country = response.text
        else:
            move.country = 'XYZ'

        response = requests.get(
            'https://geo.kumy.org/api/getElevation?lat={}&lon={}'.format(move.latitude, move.longitude))
        if response.ok and response.text != 'None':
            move.altitude = response.text
        else:
            move.altitude = '-2000'
    else:
        move.country = ''
        move.altitude = -32768


def update_geokret_total_moves_count(geokret_id):
    """ Update GeoKret total move count
    """
    moves = Move.query \
        .filter(Move.geokret_id == geokret_id) \
        .filter(Move.type.in_((MOVE_TYPE_DROPPED, MOVE_TYPE_SEEN, MOVE_TYPE_DIPPED))) \
        .order_by(Move.moved_on_datetime.desc())

    geokret = Geokret.query.get(geokret_id)
    geokret.caches_count = moves.count()


def update_geokret_holder(geokret_id):
    """ Update GeoKret holder
    """
    moves = Move.query \
        .filter(Move.geokret_id == geokret_id) \
        .filter(Move.type != MOVE_TYPE_COMMENT) \
        .order_by(Move.moved_on_datetime.desc())

    geokret = Geokret.query.get(geokret_id)

    geokret.holder_id = None
    if moves.count():
        for move in moves:
            if move.type in [MOVE_TYPE_DROPPED, MOVE_TYPE_ARCHIVED, MOVE_TYPE_SEEN]:
                break
            elif move.type in [MOVE_TYPE_GRABBED, MOVE_TYPE_DIPPED]:
                geokret.holder_id = move.author_id
                break


def update_geokret_missing(geokret):
    """ Update GeoKret missing status
    """
    if not isinstance(geokret, Geokret):
        geokret = Geokret.query.get(geokret)

    if geokret.last_position is not None:
        for comment in geokret.last_position.comments:
            if comment.type == MOVE_COMMENT_TYPE_MISSING:
                geokret.missing = True
                return
    geokret.missing = False


def update_move_comments_type(move_id):
    """ Convert move comment type to comment when necessary
    """
    move = Move.query.get(move_id)
    if move.type in (MOVE_TYPE_DIPPED, MOVE_TYPE_COMMENT, MOVE_TYPE_GRABBED, MOVE_TYPE_ARCHIVED):
        for comment in move.comments:
            comment.type = MOVE_COMMENT_TYPE_COMMENT
    # session.add(move)



# -*- coding: utf-8 -*-

from nose.tools import assert_raises
from sqlalchemy.exc import OperationalError

from geokrety_api_models.badge import Badge
from geokrety_api_models.user import User

from . import DatabaseTest


class TestBadge(DatabaseTest):
    """Test badge"""

    def test_create(self):
        badge = Badge()
        self.session.add(badge)
        with assert_raises(OperationalError):
            self.session.commit()
        self.session.rollback()

    def test_create_ok(self):
        user = User(name="username", password="password", email="user@geokrety.org")
        self.session.add(user)

        badge = Badge()
        badge.name = "some name"
        badge.author = user
        self.session.add(badge)

        self.assertEqual(badge.name, "some name")



# -*- coding: utf-8 -*-

import random
from datetime import datetime

from sqlalchemy import (TIMESTAMP, Boolean, Column, DateTime, Enum, ForeignKey,
                        Integer, String, Text, event, inspect)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound

import bleach
import characterentities

from .base import Base
from .move import Move
from .utilities.const import MOVE_TYPE_ARCHIVED, MOVE_TYPE_COMMENT
from .utilities.date import round_microseconds


class Geokret(Base):
    __tablename__ = 'gk-geokrety'

    id = Column(
        'id',
        Integer,
        primary_key=True,
        key='id',
    )
    tracking_code = Column(
        'nr',
        String(9),
        key='tracking_code',
        nullable=False,
        unique=True,
    )
    _name = Column(
        'nazwa',
        String(75),
        key='name',
        nullable=False,
    )
    _description = Column(
        'opis',
        Text(),
        key='description',
        nullable=False,
        default='',
    )
    type = Column(
        'typ',
        Enum('0', '1', '2', '3', '4'),
        key='type',
        nullable=False,
    )
    missing = Column(
        'missing',
        Boolean,
        key='missing',
        nullable=False,
        default=False,
    )
    distance = Column(
        'droga',
        Integer,
        key='distance',
        nullable=False,
        default=0,
    )
    caches_count = Column(
        'skrzynki',
        Integer,
        key='caches_count',
        nullable=False,
        default=0,
    )
    pictures_count = Column(
        'zdjecia',
        Integer,
        key='pictures_count',
        nullable=False,
        default=0,
    )
    _created_on_datetime = Column(
        'data',
        DateTime,
        nullable=False,
        key='created_on_datetime',
        default=datetime.utcnow,
    )
    updated_on_datetime = Column(
        'timestamp',
        TIMESTAMP(timezone=True),
        key='updated_on_datetime',
        default=datetime.utcnow,
        onupdate=datetime.utcnow,
    )

    owner_id = Column(
        'owner',
        Integer,
        ForeignKey('gk-users.id', name='fk_geokret_owner'),
        nullable=True,
        key='owner_id',
    )
    owner = relationship(
        "User",
        foreign_keys=[owner_id],
        backref="geokrety_owned"
    )

    holder_id = Column(
        'hands_of',
        Integer,
        ForeignKey('gk-users.id', name='fk_geokret_holder'),
        key='holder_id',
    )
    holder = relationship(
        "User",
        foreign_keys=[holder_id],
        backref="geokrety_held"
    )

    # This is used to compute the archived status
    _moves = relationship(
        'Move',
        foreign_keys="Move.geokret_id",
        lazy="dynamic",
    )

    last_position_id = Column(
        'ost_pozycja_id',
        Integer,
        ForeignKey('gk-ruchy.id', name='fk_geokret_last_position'),
        key='last_position_id',
        nullable=True,
        default=None,
    )
    last_position = relationship(
        "Move",
        foreign_keys=[last_position_id],
        backref="spotted_geokret",
        post_update=True
    )

    last_move_id = Column(
        'ost_log_id',
        Integer,
        ForeignKey('gk-ruchy.id', name='fk_last_move', ondelete="SET NULL"),
        key='last_move_id'
    )
    last_move = relationship(
        "Move",
        foreign_keys=[last_move_id],
        backref="moves",
        post_update=True
    )

    # avatar_id = Column(
    #     'avatarid',
    #     Integer,
    #     ForeignKey('gk-obrazki.id'),
    #     key='avatar_id'
    # )

    @hybrid_property
    def archived(self):
        """Compute the archived status
        """
        try:
            last_move = self._moves.filter(Move.type != MOVE_TYPE_COMMENT)\
                .order_by(Move.moved_on_datetime.desc()).limit(1).one()
            return last_move.type == MOVE_TYPE_ARCHIVED
        except NoResultFound:
            return False

    @property
    def average_rating(self):
        # TODO note should come from database
        return 0

    @hybrid_property
    def name(self):
        return characterentities.decode(self._name)

    @name.setter
    def name(self, name):
        # Drop all html tags
        name_clean = bleach.clean(name, tags=[], strip=True)
        # Strip spaces
        self._name = characterentities.decode(name_clean).strip()

    @name.expression
    def name(cls):
        return cls._name

    @hybrid_property
    def description(self):
        return characterentities.decode(self._description)

    @description.setter
    def description(self, description):
        if description is None:
            description = ''
        # Drop all unallowed html tags
        description_clean = bleach.clean(description, strip=True)
        # Strip spaces
        self._description = characterentities.decode(description_clean).strip()

    @description.expression
    def description(cls):
        return cls._description

    @hybrid_property
    def created_on_datetime(self):
        if self._created_on_datetime is None:
            raise AssertionError("created_on_datetime is missing")
        if isinstance(self._created_on_datetime, str):
            self._created_on_datetime = datetime.strptime(self._created_on_datetime, "%Y-%m-%dT%H:%M:%S")
        return round_microseconds(self._created_on_datetime)

    @created_on_datetime.setter
    def created_on_datetime(self, created_on_datetime):
        self._created_on_datetime = created_on_datetime

    @created_on_datetime.expression
    def created_on_datetime(cls):
        return cls._created_on_datetime


@event.listens_for(Geokret, 'init')
def receive_init(target, args, kwargs):
    target.tracking_code = ''.join(random.choice('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ') for i in range(6))  # TODO


@event.listens_for(Geokret, 'before_insert')
def before_insert_listener(mapper, connection, target):
    if not target._created_on_datetime:
        target._created_on_datetime = round_microseconds(datetime.utcnow())
    target.updated_on_datetime = target._created_on_datetime


@event.listens_for(Geokret, 'before_update')
def before_update_listener(mapper, connection, target):
    target.updated_on_datetime = round_microseconds(datetime.utcnow())


MONITORED_ATTRIBUTES = [
    'tracking_code',
    '_name',
    '_description',
    'type',
    'missing',
    'distance',
    'caches_count',
    'pictures_count',
    'updated_on_datetime',
    'owner_id',
    'holder_id',
    'last_position_id',
    'last_move_id',
]


def _has_changes_that_need_event(instance):
    instance_attrs = inspect(instance).attrs
    for attribute in MONITORED_ATTRIBUTES:
        if hasattr(instance_attrs, attribute) and \
                getattr(instance_attrs, attribute).history.has_changes():
            return True



# -*- coding: utf-8 -*-

import datetime

from sqlalchemy import (Column, Date, DateTime, ForeignKey,
                        ForeignKeyConstraint, Integer)
from sqlalchemy.ext.hybrid import hybrid_property

from .base import Base


class NewsSubscription(Base):
    __tablename__ = 'gk-news-comments-access'

    id = Column(
        Integer,
        primary_key=True,
    )

    news_id = Column(
        Integer,
        ForeignKey('gk-news.id'),
        nullable=False,
    )

    user_id = Column(
        Integer,
        ForeignKey('gk-users.id', name='fk_news_comment_subscription'),
        nullable=False,
    )

    ForeignKeyConstraint(
        ['author_id'], ['user.id'],
        use_alter=True,
        name='fk_news_comment_subscription'
    )

    subscribed_on_datetime = Column(
        'read',
        DateTime,
        key='subscribed_on_datetime',
        nullable=False,
        default=datetime.datetime.utcnow,
    )

    post = Column(
        Date,
        nullable=True,
    )

    _subscribed = Column(
        'subscribed',
        Integer,
        key='subscribed',
        default=1,
    )

    @hybrid_property
    def subscribed(self):
        """
        Hybrid property for subscribed
        :return:
        """
        return bool(self._subscribed)

    @subscribed.setter
    def subscribed(self, subscribed):
        """
        Setter for _subscribed, convert integer to boolean
        :param subscribed:
        :return:
        """
        self._subscribed = bool(subscribed)



# -*- coding: utf-8 -*-

import datetime
import uuid

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, event
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import bleach
import characterentities

from .base import Base


class Badge(Base):
    __tablename__ = 'gk-badges-collection'

    id = Column(
        'id',
        Integer,
        primary_key=True,
        key='id',
    )

    _name = Column(
        'name',
        String(80),
        key='name',
        nullable=False,
        unique=True,
    )

    _description = Column(
        'description',
        String(128),
        key='description',
        nullable=True,
    )

    filename = Column(
        'filename',
        String(32),
        key='filename',
        nullable=False,
    )

    created_on_datetime = Column(
        'date',
        DateTime,
        key='created_on_datetime',
        nullable=False,
        default=datetime.datetime.utcnow,
    )

    author_id = Column(
        'author_id',
        Integer,
        ForeignKey('gk-users.id'),
        key='author_id',
        nullable=False,
    )

    author = relationship(
        "User",
        foreign_keys=[author_id],
        backref="created_badges"
    )

    @hybrid_property
    def name(self):
        return characterentities.decode(self._name)

    @name.setter
    def name(self, name):
        name_clean = bleach.clean(name, strip=True)
        self._name = characterentities.decode(name_clean).strip()

    @name.expression
    def name(cls):
        return cls._name

    @hybrid_property
    def description(self):
        if self._description is None:
            return u''
        return characterentities.decode(self._description)

    @description.setter
    def description(self, description):
        if description is None:
            return u''
        description_clean = bleach.clean(description, strip=True)
        self._description = characterentities.decode(description_clean).replace('\x00', '').strip()

    @description.expression
    def description(cls):
        return cls._description

    @hybrid_property
    def upload_url(self):
        # Todo send info to other daemon -> redis?
        from app.views.minio import storage
        res = storage.connection.presigned_put_object(
            'badges-incoming', self.filename)  # , expires=timedelta(minutes=5))
        return res
        # return uuid.uuid4().hex


@event.listens_for(Badge, 'init')
def receive_init(target, args, kwargs):
    target.filename = uuid.uuid4().hex


if __name__ == '__main__':
    # Check
    badge = Badge()



# -*- coding: utf-8 -*-

from nose.tools import assert_raises

from geokrety_api_exceptions import GKUnprocessableEntity
from geokrety_api_models import Geokret, Move, User
from geokrety_api_models.utilities.const import (GEOKRET_TYPE_TRADITIONAL,
                                                 MOVE_TYPE_GRABBED)

from . import DatabaseTest


class TestMove(DatabaseTest):
    """Test move"""

    def test_create(self):
        move = Move(moved_on_datetime="2019-01-19T22:46:44")
        self.session.add(move)
        with assert_raises(GKUnprocessableEntity):
            self.session.commit()

    def test_create_ok(self):
        user = User(name="username", password="password", email="user@geokrety.org")
        geokret = Geokret(name="My GeoKret",
                          type=GEOKRET_TYPE_TRADITIONAL,
                          created_on_datetime="2019-01-19T22:46:44")
        move1 = Move(geokret=geokret,
                     moved_on_datetime="2019-01-19T22:46:44",
                     type=MOVE_TYPE_GRABBED,
                     author=user)

        self.session.add(move1)
        self.session.commit()

        move2 = Move(geokret=geokret,
                     moved_on_datetime="2019-01-19T22:46:44",
                     type=MOVE_TYPE_GRABBED,
                     author=user)

        self.session.add(move2)
        with assert_raises(GKUnprocessableEntity):
            self.session.commit()



# -*- coding: utf-8 -*-

import datetime

from sqlalchemy import (Column, DateTime, ForeignKey,
                        ForeignKeyConstraint, Integer, String)
from sqlalchemy.ext.hybrid import hybrid_property

import bleach
import characterentities

from .base import Base


class NewsComment(Base):
    __tablename__ = 'gk-news-comments'

    id = Column(
        'comment_id',
        Integer,
        primary_key=True,
        key='id'
    )

    news_id = Column(
        'news_id',
        Integer,
        ForeignKey('gk-news.id'),
        key='news_id',
        nullable=False
    )

    author_id = Column(
        'user_id',
        Integer,
        ForeignKey('gk-users.id', name='fk_news_comment_author'),
        key='author_id',
        nullable=False
    )

    ForeignKeyConstraint(
        ['author_id'], ['user.id'],
        use_alter=True,
        name='fk_news_comment_author'
    )

    created_on_datetime = Column(
        'date',
        DateTime,
        key='created_on_datetime',
        nullable=False,
        default=datetime.datetime.utcnow
    )

    _comment = Column(
        'comment',
        String(1000),
        nullable=False
    )

    icon = Column(
        Integer,
        default=0
    )

    # news = relationship('News', backref=backref('news_comments'))
    # author = relationship('User', backref=backref('news_comments'))

    @hybrid_property
    def comment(self):
        return characterentities.decode(self._comment)

    @comment.setter
    def comment(self, comment):
        # Drop all html tags
        comment_clean = bleach.clean(comment, strip=True)
        # Strip spaces
        self._comment = characterentities.decode(comment_clean).strip()

    @comment.expression
    def comment(cls):
        return cls._comment



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from sqlalchemy import (Column, DateTime, Enum, ForeignKey, Integer, String,
                        event, func, inspect, select)
from sqlalchemy.dialects.mysql import DOUBLE
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship

import bleach
import characterentities
from geokrety_api_exceptions.json_api import GKUnprocessableEntity

from .base import Base
from .utilities.date import round_microseconds


class Move(Base):
    __tablename__ = 'gk-ruchy'

    id = Column(
        'ruch_id',
        Integer,
        primary_key=True,
        key='id'
    )

    geokret_id = Column(
        'id',
        Integer,
        ForeignKey('gk-geokrety.id', name='fk_geokret_moved'),
        key='geokret_id',
        nullable=False,
        default=None
    )
    geokret = relationship(
        "Geokret",
        foreign_keys=[geokret_id],
        backref=backref("moves", order_by="-Move.moved_on_datetime")
    )

    latitude = Column(
        'lat',
        DOUBLE(precision=8, scale=5),
        key='latitude',
        nullable=True,
        default=None
    )

    longitude = Column(
        'lon',
        DOUBLE(precision=8, scale=5),
        key='longitude',
        nullable=True,
        default=None
    )

    altitude = Column(
        'alt',
        Integer,
        key='altitude',
        nullable=True,
        default=None,
    )

    country = Column(
        'country',
        String(3),
        key='country',
        nullable=True,
        default=None,
    )

    distance = Column(
        'droga',
        Integer,
        key='distance',
        nullable=False,
        default=0
    )

    waypoint = Column(
        'waypoint',
        String(10),
        key='waypoint',
        nullable=False,
        default=''
    )

    _comment = Column(
        'koment',
        String(5120),
        key='comment',
        nullable=False,
        default=''
    )

    pictures_count = Column(
        'zdjecia',
        Integer,
        key='pictures_count',
        nullable=False,
        default=0
    )

    comments_count = Column(
        'komentarze',
        Integer,
        key='comments_count',
        nullable=False,
        default=0
    )

    type = Column(
        'logtype',
        Enum('0', '1', '2', '3', '4', '5'),
        key='type',
        nullable=False,
    )

    author_id = Column(
        'user',
        Integer,
        ForeignKey('gk-users.id'),
        key='author_id',
        nullable=False,
        default=None,
    )
    author = relationship("User", foreign_keys=[author_id], backref="moves")

    username = Column(
        'username',
        String(20),
        key='username',
        nullable=False,
        default=''
    )

    # This is used to compute the missing status
    _move_comments = relationship(
        'MoveComment',
        foreign_keys="MoveComment.move_id",
        lazy="dynamic",
    )

    moved_on_datetime = Column(
        'data',
        DateTime,
        key='moved_on_datetime',
        nullable=False,
        default=datetime.utcnow,
    )

    created_on_datetime = Column(
        'data_dodania',
        DateTime,
        key='created_on_datetime',
        nullable=False,
        default=datetime.utcnow,
    )

    updated_on_datetime = Column(
        'timestamp',
        DateTime,
        key='updated_on_datetime',
        default=datetime.utcnow,
        onupdate=datetime.utcnow
    )

    _application_name = Column(
        'app',
        String(16),
        key='application_name',
        nullable=True
    )

    _application_version = Column(
        'app_ver',
        String(16),
        key='application_version',
        nullable=True
    )

    @hybrid_property
    def comment(self):
        return characterentities.decode(self._comment)

    @comment.setter
    def comment(self, comment):
        comment_clean = bleach.clean(comment, strip=True)
        self._comment = characterentities.decode(comment_clean).strip()

    @comment.expression
    def comment(cls):
        return cls._comment

    @hybrid_property
    def application_version(self):
        if self._application_version is None:
            return None
        return characterentities.decode(self._application_version)

    @application_version.setter
    def application_version(self, application_version):
        if application_version is None:
            self._application_version = None
        else:
            application_version_clean = bleach.clean(application_version, tags=[], strip=True)
            self._application_version = characterentities.decode(application_version_clean).strip()

    @application_version.expression
    def application_version(cls):
        return cls._application_version

    @hybrid_property
    def application_name(self):
        if self._application_name is None:
            return None
        return characterentities.decode(self._application_name)

    @application_name.setter
    def application_name(self, application_name):
        if application_name is None:
            self._application_name = None
        else:
            application_name_clean = bleach.clean(application_name, tags=[], strip=True)
            self._application_name = characterentities.decode(application_name_clean).strip()

    @application_name.expression
    def application_name(cls):
        return cls._application_name

    @hybrid_property
    def _moved_on_datetime(self):
        if self.moved_on_datetime is None:
            raise AssertionError("moved_on_datetime is missing")
        if isinstance(self.moved_on_datetime, str):
            self.moved_on_datetime = datetime.strptime(self.moved_on_datetime, "%Y-%m-%dT%H:%M:%S")
        return round_microseconds(self.moved_on_datetime)


@event.listens_for(Move, 'before_update')
@event.listens_for(Move, 'before_insert')
def my_before_insert_listener(mapper, connection, target):
    if target.geokret is None:
        raise GKUnprocessableEntity("Move must concern a GeoKret",
                                    {'pointer': '/data/relationships/geokret'})

    # Move cannot be done before GeoKret birth
    if target._moved_on_datetime < target.geokret.created_on_datetime:
        raise GKUnprocessableEntity("Move date cannot be prior GeoKret birth date",
                                    {'pointer': '/data/attributes/moved-on-datetime'})

    # Move cannot be done in the future
    if target._moved_on_datetime > datetime.utcnow().replace(microsecond=0) + timedelta(seconds=1):
        raise GKUnprocessableEntity("Move date cannot be in the future",
                                    {'pointer': '/data/attributes/moved-on-datetime'})

    # Identical move date is forbidden
    if _has_changes_that_need_recompute(target):
        move_table = Move.__table__
        res = connection.execute(
            select([func.count()]).select_from(move_table).
            where(Move.moved_on_datetime == target.moved_on_datetime).
            where(Move.geokret_id == target.geokret.id).
            where(Move.id != target.id)
        ).scalar()
        if res > 0:
            raise GKUnprocessableEntity("There is already a move at that time",
                                        {'pointer': '/data/attributes/moved-on-datetime'})


def _has_changes_that_need_recompute(instance):
    if inspect(instance).attrs.type.history.has_changes() or \
            inspect(instance).attrs.moved_on_datetime.history.has_changes() or \
            inspect(instance).attrs.geokret.history.has_changes():
        return True



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="geokrety_api_models",
    version="0.0.2",
    author="Mathieu Alorent",
    author_email="kumy@geokrety.org",
    description="The GeoKrety API ORM",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/geokrety/geokrety-api-models",
    packages=setuptools.find_packages(),
    install_requires=[
        "bleach==3.1.0",
        "characterentities==0.1.2",
        "SQLAlchemy==1.2.13",
        "mysqlclient==1.3.14",
        "bcrypt==3.1.6",
        "geopy==1.18.1",
        "requests[security]==2.21.0",
        "geokrety-api-exceptions==0.0.1",
    ],
    dependency_links=[
        "git+https://github.com/exavolt/python-phpass.git#egg=python-phpass",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# -*- coding: utf-8 -*-

from datetime import timedelta


def round_microseconds(date):
    fraction = date.microsecond / 1000000.0
    rounded = round(fraction, 0)

    if rounded >= 1:
        # round up by adding a second to the datetime
        date = date.replace(microsecond=0) + timedelta(seconds=1)

    return date



# -*- coding: utf-8 -*-

from .badge import Badge
from .geokret import Geokret
from .move import Move
from .move_comment import MoveComment
from .news import News
from .news_comment import NewsComment
from .news_subscription import NewsSubscription
from .user import User

__all__ = [
    "Badge",
    "Geokret",
    "Move",
    "MoveComment",
    "News",
    "NewsComment",
    "NewsSubscription",
    "User",
]



# -*- coding: utf-8 -*-

GEOKRET_TYPE_TRADITIONAL = "0"
GEOKRET_TYPE_BOOK = "1"
GEOKRET_TYPE_HUMAN = "2"
GEOKRET_TYPE_COIN = "3"
GEOKRET_TYPE_KRETYPOST = "4"

GEOKRET_TYPES_TEXT = {
    GEOKRET_TYPE_TRADITIONAL: "Traditional",
    GEOKRET_TYPE_BOOK: "A book/CD/DVD",
    GEOKRET_TYPE_HUMAN: "A Human",
    GEOKRET_TYPE_COIN: "A coin",
    GEOKRET_TYPE_KRETYPOST: "KretyPost",
}

GEOKRETY_TYPES = [
    {'id': GEOKRET_TYPE_TRADITIONAL, 'name': GEOKRET_TYPES_TEXT[GEOKRET_TYPE_TRADITIONAL]},
    {'id': GEOKRET_TYPE_BOOK, 'name': GEOKRET_TYPES_TEXT[GEOKRET_TYPE_BOOK]},
    {'id': GEOKRET_TYPE_HUMAN, 'name': GEOKRET_TYPES_TEXT[GEOKRET_TYPE_HUMAN]},
    {'id': GEOKRET_TYPE_COIN, 'name': GEOKRET_TYPES_TEXT[GEOKRET_TYPE_COIN]},
    {'id': GEOKRET_TYPE_KRETYPOST, 'name': GEOKRET_TYPES_TEXT[GEOKRET_TYPE_KRETYPOST]},
]

GEOKRETY_TYPES_LIST = [
    GEOKRET_TYPE_TRADITIONAL,
    GEOKRET_TYPE_BOOK,
    GEOKRET_TYPE_HUMAN,
    GEOKRET_TYPE_COIN,
    GEOKRET_TYPE_KRETYPOST,
]

MOVE_TYPE_DROPPED = "0"
MOVE_TYPE_GRABBED = "1"
MOVE_TYPE_COMMENT = "2"
MOVE_TYPE_SEEN = "3"
MOVE_TYPE_ARCHIVED = "4"
MOVE_TYPE_DIPPED = "5"

MOVE_TYPES_TEXT = {
    MOVE_TYPE_DROPPED: 'Dropped to',
    MOVE_TYPE_GRABBED: 'Grabbed from',
    MOVE_TYPE_COMMENT: 'A comment',
    MOVE_TYPE_SEEN: 'Seen in',
    MOVE_TYPE_ARCHIVED: 'Archived',
    MOVE_TYPE_DIPPED: 'Dipped',
}

MOVE_TYPES = [
    {'id': MOVE_TYPE_DROPPED, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_DROPPED]},
    {'id': MOVE_TYPE_GRABBED, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_GRABBED]},
    {'id': MOVE_TYPE_COMMENT, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_COMMENT]},
    {'id': MOVE_TYPE_SEEN, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_SEEN]},
    {'id': MOVE_TYPE_ARCHIVED, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_ARCHIVED]},
    {'id': MOVE_TYPE_DIPPED, 'name': MOVE_TYPES_TEXT[MOVE_TYPE_DIPPED]},
]

MOVE_TYPES_LIST = [
    MOVE_TYPE_DROPPED,
    MOVE_TYPE_GRABBED,
    MOVE_TYPE_COMMENT,
    MOVE_TYPE_SEEN,
    MOVE_TYPE_ARCHIVED,
    MOVE_TYPE_DIPPED,
]
# MOVE_TYPES_COUNT = 6

MOVE_COMMENT_TYPE_COMMENT = 0
MOVE_COMMENT_TYPE_MISSING = 1

MOVE_COMMENT_TYPES_LIST = [
    MOVE_COMMENT_TYPE_COMMENT,
    MOVE_COMMENT_TYPE_MISSING,
]

LANGUAGES = {
    'fr': u'France',
    'pl': u'Polska',
    'de': u'Deutschland',
    'en': u'English',
    'ru': u'Россия',
    'ro': u'România',
}

COUNTRIES = {
    'fr': u'France',
    'pl': u'Polska',
    'de': u'Deutschland',
    'uk': u'England',
    'ru': u'Россия',
    'ro': u'România',
}
